import json
import re
import warnings
from importlib import import_module

import click
from prance import ResolvingParser
from yaml import FullLoader, load

from apievaluator.logic import eval_all


@click.command()
@click.option(
    '-s', '--spec',
    help='Path to the spec file',
    type=click.File('r'), required=True
)
@click.option(
    '-u', '--apiurl',
    help='Base URL of the API',
    required=True
)
@click.option(
    '-t', '--tests',
    help='Path to the tests file',
    type=click.File('r'), required=True
)
@click.option(
    '-o', '--output', 'o',
    help='If specified, write to file instead of stdout',
    type=click.File('w'), default='-'
)
def main(spec: click.File, apiurl: str, tests: click.File, o: click.File):

    try:
        openApi = ResolvingParser(spec.name)
        spec = openApi.specification
        tests = load(tests, Loader=FullLoader)
    except Exception as err:
        click.echo('Error parsing input file: %s' % err)
        return

    [exists, resposes, testsDict] = eval_all(spec['paths'], apiurl, tests)

    click.echo('Summary:', file=o)
    click.echo('\tFailed:  %3d' % testsDict["failed"], file=o)
    click.echo('\tSuccess: %3d' % testsDict["success"], file=o)
    click.echo('\tTotal:   %3d' % testsDict["test"], file=o)

    click.echo('Route present checks:', file=o)
    for result in exists:
        click.echo('\t%s' % result[1], file=o)

    click.echo('Response checks:', file=o)
    for result in resposes:
        ret = result[1]

        if isinstance(ret, str):
            click.echo('\t%s' % ret, file=o)
        else:
            click.echo('\t%s' % ret[0], file=o)
            for tail in ret[1:]:
                click.echo('\t\t%s' % tail[1], file=o)


if __name__ == "__main__":
    main()



from apievaluator.custom_tests import AbstractTest


class Boolean(AbstractTest):

    def __init__(self, name: str, should_be: bool):
        self.name = name
        self.should_be = should_be

    def run(self, data: dict):
        res = data[self.name] == self.should_be
        return (res, 'OK' if res else 'Property %s was %s instead of %s' % (self.name, data[self.name], self.should_be))



from apievaluator.custom_tests import AbstractTest


class ArrayOfLength(AbstractTest):

    def __init__(self, name: str, len: int):
        self.name = name
        self.len = len

    def run(self, data: dict):
        val = data[self.name]

        if isinstance(val, list):
            if len(val) == self.len:
                return (True, 'OK')
            else:
                return (False, '%s is of length %d, expected %d', self.name, len(val), self.len)
        else:
            return (False, '%s is not an array, got %s instead', self.name, type(val).__name__)



from apievaluator.custom_tests import AbstractTest


class Positive(AbstractTest):

    def __init__(self, name: str):
        self.name = name

    def run(self, data: dict):
        val = data[self.name]

        if isinstance(val, (int, float)):
            return (val >= 0, 'OK - %s is positive' % self.name if val >= 0 else 'Expected positive number, got %d' % val)
        elif isinstance(val, str) and val.isnumeric():
            return (False, 'Expected a number, got numeric string')
        else:
            return (False, 'Expected a number, got %s' % type(val).__name__)



from apievaluator.custom_tests import AbstractTest


class ObjectWithKeys(AbstractTest):

    def __init__(self, name: str, with_keys=[]):
        self.name = name
        self.with_keys = with_keys

    def run(self, data: dict):
        if isinstance(data[self.name], dict):

            prop = data[self.name]
            errors = []

            for key in self.with_keys:
                if key not in prop:
                    errors.append(key)

            if not errors:
                return (True, 'OK')
            else:
                return (False, 'Object is missing keys: %s' % errors)
        else:
            type = type(data[self.name]).__name__
            return (False, 'Property \'%s\' was of type %s, expected an object' % (self.name, type))



from setuptools import setup, find_packages

setup(
    name='apievaluator',
    author='Filip Majetić',
    author_email='filip.majetic@fer.hr',
    version='0.0.1',
    description='Validate an API specification and run tests',
    keywords=['rest', 'api', 'testing', 'automated'],
    url='https://gitlab.com/fmajestic/api-evaluator',
    download_url='https://gitlab.com/fmajestic/api-evaluator/-/archive/v0.1.1/api-evaluator-v0.1.1.tar.gz',
    packages=find_packages(),
    include_package_data=True,
    install_requires=['click',
                      'prance',
                      'pyyaml',
                      'requests',
                      ],
    entry_points={
        'console_scripts': ['apieval=apievaluator.main:main']
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
)



import json
import re
from importlib import import_module

import click
import requests


def checkObjectTypes(resp, cond, testDict):
    ret = []
    if isinstance(cond, str):
        testDict['test'] += 1
        if cond == 'int':
            if isinstance(resp, int):
                testDict['success'] = testDict['success']+1
                ret.append((True, 'OK - int'))
            else:
                testDict['failed'] = testDict['failed']+1
                ret.append((False, 'Not OK - expected %s, got %s' % (cond, type(resp).__name__)))
        elif cond == 'string':
            if isinstance(resp, str):

                testDict['success'] = testDict['success']+1
                ret.append((True, 'OK - string'))
            else:
                testDict['failed'] = testDict['failed']+1
                ret.append((False, 'Not OK - expected %s, got %s' % (cond, type(resp).__name__)))
        elif cond == 'array':
            if isinstance(resp, list):

                testDict['success'] = testDict['success']+1
                ret.append((True, 'OK - array'))
            else:
                testDict['failed'] = testDict['failed']+1
                ret.append((False, 'Not OK - expected %s, got %s' % (cond, type(resp).__name__)))

    elif isinstance(cond, (object, dict)):
        for key, dataType in cond.items():
            testDict['test'] = testDict['test']+1
            if dataType == 'int':
                if isinstance(resp.get(key), int):
                    testDict['success'] = testDict['success']+1
                    ret.append((True, 'OK - int'))
                else:
                    testDict['failed'] = testDict['failed']+1
                    ret.append((False, 'Not OK - expected %s, got %s' % (cond, type(resp).__name__)))
            elif dataType == 'string':
                if isinstance(resp.get(key), str):
                    testDict['success'] = testDict['success']+1
                    ret.append((True, 'OK - string'))
                else:
                    testDict['failed'] = testDict['failed']+1
                    ret.append((False, 'Not OK - expected %s, got %s' % (cond, type(resp).__name__)))
            elif dataType == 'array':
                if isinstance(resp.get(key), list):
                    testDict['success'] = testDict['success']+1
                    ret.append((True, 'OK - array'))
                else:
                    testDict['failed'] = testDict['failed']+1
                    ret.append((False, 'Not OK - expected %s, got %s' % (cond, type(resp).__name__)))
            elif isinstance(dataType, object):
                testDict['success'] = testDict['success']+1
                ret.extend(checkObjectTypes(resp.get(key), dataType, testDict))
            else:
                testDict['failed'] = testDict['failed']+1
    return ret


def load_custom_test(name: str, prop: str, arg: str):
    kls = getattr(import_module('ApiEvaluator.CustomTests'), name)
    return kls(prop, arg) if arg else kls(prop)


def CheckIfEndpointsExist(paths, apiurl):
    result = {}
    routes = []
    ret = []
    testDict = {
        'total': 0,
        'success': 0,
        'failed': 0
    }

    for key in paths:
        routes.append(apiurl+key)
    for r in routes:
        TestGet(routes, result, r)

    for key in result:
        testDict['test'] = testDict['test']+1
        if result[key] == True:
            testDict['success'] = testDict['success']+1
            ret.append((True, '%s - Exists' % key))
        elif result[key] == False:
            testDict['failed'] = testDict['failed']+1
            ret.append((False, '%s - Does not exist' % key))

    return (ret, testDict)


def TestGet(routes, result, r):
    res = requests.get(r, verify=False)

    # result[r] = res.status_code >= 200 and res.status_code < 400

    if str(res.status_code)[0] == '2':
        result[r] = True
    elif res.status_code == 404:
        result[r] = False
    elif res.reason != 'Not Found':
        result[r] = True
    else:
        result[r] = True


def EvaluateResponses(apiurl, tests):
    endpoints = tests["endpoints"]
    endpointSpec = {}
    resource = {}
    ret = []
    testDict = {
        'total': 0,
        'success': 0,
        'failed': 0
    }

    for key, endpointSpec in endpoints.items():
        endpointMethods = endpointSpec.get('data')
        for method, condition in endpointMethods.items():
            results = []
            results.append('%s %s' % (method, apiurl + key))

            if method == 'GET':
                expectedType = condition.get('resp').get('type')
                url = apiurl

                if condition.get('req') == None:
                    url = apiurl+key
                elif condition.get('req').get('in') == 'route':
                    a = re.sub(r'\{.*\}', condition.get('req').get('Id'), key)
                    url = apiurl+a

                r = requests.get(url, verify=False)
                resp = json.loads(r.text)
                if expectedType != "object":
                    results.extend(checkObjectTypes(resp, condition.get('resp').get('type'), testDict))
                else:
                    results.extend(checkObjectTypes(resp, condition.get('resp').get('object'), testDict))

                custom = condition.get('custom')
                if custom:
                    for test in custom:
                        t = load_custom_test(test['name'], test['prop'], test.get('arg'), import_module)
                        results.append(t.run(resp))
            elif method == 'POST':
                if condition.get('req').get('in') != "body":
                    results.append((False, f'No body data for POST {key}'))
                else:
                    url = apiurl+key
                    reqData = condition.get('req').get('object')
                    r1 = requests.post(url, json=reqData, verify=False)
                    if r1.status_code == 201:
                        click.secho(str(r1.status_code) + " Created Succesfuly", bg="green", fg="white")
                        results.extend(checkObjectTypes(json.loads(r1.text), condition.get('resp').get('object'), testDict))
                    else:
                        click.secho("Unexpected Error", bg='red', fg='white')
                        testDict['test'] = testDict['test']+1
                        testDict['failed'] = testDict['failed']+1
                        results.append((False, 'Unexpected error, 1st attempt status code %d' % r1.status_code))

                    r2 = requests.post(url, json=reqData, verify=False)
                    if r2.status_code == 400:
                        click.secho(str(r2.status_code)+" Bad request: \"" +
                                    r2.text+"\" Expected", bg="green", fg="white")
                    else:
                        testDict['test'] = testDict['test']+1
                        testDict['failed'] = testDict['failed']+1
                        results.append((False, 'Unexpected error, 2nd attempt status code %d' % r2.status_code))

                    custom = condition.get('custom')
                    if custom:
                        for test in custom:
                            t = load_custom_test(test['name'], test['prop'], test.get('arg'), import_module)
                            results.append(t.run(r1.json()))
            elif method == 'PUT':
                if condition.get('req').get('in') != "both":
                    results.append((False, f'No body data for POST {key}'))
                else:
                    a = re.sub(r'\{.*\}', condition.get('req').get('Id'), key)
                    url = apiurl+a
                    reqData = condition.get('req').get('object')
                    r1 = requests.put(url, json=reqData, verify=False)
                    if r1.status_code == 200:
                        results.extend(checkObjectTypes(json.loads(r1.text), condition.get('resp').get('object'), testDict))
                    else:
                        results.append((False, f'Bad status code for PUT {key} -> {r1.status_code}'))
                    r2 = requests.put(url, json=reqData, verify=False)
                    if r2.status_code == 400:
                        click.secho(str(r2.status_code)+" Bad request: \"" + r2.text+"\" Expected", bg="green", fg="white")
                    else:
                        testDict['test'] = testDict['test']+1
                        testDict['failed'] = testDict['failed']+1

            ret.append((True, results))

    return (ret, testDict)


def merge(A: dict, B: dict, f: lambda x, y: (x, y)):
    merged = {k: A.get(k, B.get(k)) for k in A.keys() ^ B.keys()}
    merged.update({k: f(A[k], B[k]) for k in A.keys() & B.keys()})
    return merged


def eval_all(paths, apiurl, tests):
    [exist, count1] = CheckIfEndpointsExist(paths, apiurl)
    [tests, count2] = EvaluateResponses(apiurl, tests)
    return (exist, tests, merge(count1, count2, lambda x, y: x+y))



from .AbstractTest import AbstractTest
from .ArrayOfLength import ArrayOfLength
from .Boolean import Boolean
from .ObjectWithKeys import ObjectWithKeys
from .Positive import Positive
from .StringRegex import StringRegex



import re

from apievaluator.custom_tests import AbstractTest


class StringRegex(AbstractTest):

    def __init__(self, name: str, regex: str):
        self.name = name
        self.regex = regex

    def run(self, data: dict):
        match = re.search(self.regex, data[self.name])

        if match:
            return (True, 'OK - regex %s matched' % self.regex)
        else:
            return (False, 'Regex %s failed for string %s' % (self.regex, data[self.name]))
